ezInterrupt = LibStub("AceAddon-3.0"):NewAddon("ezInterrupt", "AceConsole-3.0", "AceEvent-3.0")

-----------------------------
-- experimental code
-- used for removing spellIDs from the blackList/whiteList
function tRemoveByValue(table, value)
	for i,v in ipairs(table) do 
		if v == value then tremove(table, i) end
	end
end

-- used for debugging
-- it's possible to add the same spell id multiple times - don't allow that
function showBlacklistItems()
	print("ezInterrupt: the blacklist contains:")
	for i,v in ipairs(ezInterrupt.db.profile.blackList) do 
		print(i .. ": " .. v .. " (" .. select(1, GetSpellInfo(v)) ..")");
	end
end
-- used for debugging
function showWhitelistItems()
	print("ezInterrupt: the whitelist contains:")
	for i,v in ipairs(ezInterrupt.db.profile.whiteList) do 
		print(i .. ": " .. v .. " (" .. select(1, GetSpellInfo(v)) ..")");
	end
end

-- experimental code end
-----------------------------

function UpdateFilterList(table, spellID) -- adds/removes spell ID's to a blacklist or whitelist table
	local spellName = select(1, GetSpellInfo(spellID))
	if tContains(table, spellID) then
		if spellName == nil then spellName = "Invalid Spell" end
		for i,v in ipairs(table) do 
			if v == spellID then tremove(table, i) end
			if table == ezInterrupt.db.profile.whiteList then
				print("ezInterrupt: removed " .. spellID .. " (" .. spellName ..") from the whitelist.")
			elseif table == ezInterrupt.db.profile.blackList then
				print("ezInterrupt: removed " .. spellID .. " (" .. spellName ..") from the blacklist.")
			end
		end
	elseif spellName == nil then
		print("ezInterrupt: invalid spell ID")
	else
		tinsert(table, spellID)
		if table == ezInterrupt.db.profile.whiteList then
			print("ezInterrupt: added " .. spellID .. " (" .. spellName ..") to the whitelist.")
		elseif table == ezInterrupt.db.profile.blackList then
			print("ezInterrupt: added " .. spellID .. " (" .. spellName ..") to the blacklist.")
		end
	end	
end


local defaults = {
    profile = {
        enableAddon = true,
		enableAnnouncing = true,
		enableWarning = true,
		enableWarningSound = true,
		enableWarningText = true,
		enableWhispering = true,
		enableAnnouncingForFailedInterrupts = true,
		enableCustomChannel = true,
		whisperTarget = nil,
		customChannel = nil,
		slashSay = true,
		slashParty = true,
		slashRaid = false,
		useInRaid = true,
		useInParty = true,
		useAlways = true,
		soundFile = "Sound\\Spells\\ShaysBell.ogg",
		filterMode = 1, -- 2 = filter by whiteList, 1 = filter by blackList, 0 = don't filter 
		whiteList = {},
		blackList = {},
		warningTextScale = 1,
		warningTextAlpha = 1,
		warningTextOffsetX = 0,
		warningTextOffsetY = 0,
    },
}

local init = false
local outputChannel = nil
local playerInterrupts = {} -- this lists the interrupts that the player's class can provide. It is used to determine failed interrupts
local filterModeOptions = {
	[1] = "All spells trigger warnings",
	[2] = "Only disable warnings for spells in the blacklist",
	[3] = "Only enable warnings for spells in the whitelist",
}

local options = {
    name = "ezInterrupt Configuration",
    handler = ezInterrupt,
    type = "group",
    args = {
		enable = {
			type = "toggle",
			name = "Enable Addon",
			width = "full",
			set = function(info, val) ezInterrupt.db.profile.enableAddon = val if ezInterrupt.db.profile.enableAddon then ezInterrupt:OnEnable() else ezInterrupt:OnDisable() end end,
			get = function(info) return ezInterrupt.db.profile.enableAddon end,
			order = 1
		},
		announcing = {
			type = "toggle",
			name = "Announce successful interrupts",
			width = "full",
			set = function(info, val) ezInterrupt.db.profile.enableAnnouncing = val end,
			get = function(info) return ezInterrupt.db.profile.enableAnnouncing end,
			order = 2
		},
		announcingfailed = {
			type = "toggle",
			name = "Announce failed interrupts",
			width = "full",
			set = function(info, val) ezInterrupt.db.profile.enableAnnouncingForFailedInterrupts = val end,
			get = function(info) return ezInterrupt.db.profile.enableAnnouncingForFailedInterrupts end,
			order = 3
		},
		warning = {
			type = "toggle",
			name = "Enable warnings",
			width = "full",
			set = function(info, val) ezInterrupt.db.profile.enableWarning = val end,
			get = function(info) return ezInterrupt.db.profile.enableWarning end,
			order = 4
		},
		output = {
			type = "group",
			name = "Announcement Options",
			order = 5,
			inline = true,
			args = {
				header1 = {
					type = "description",
					name = "Check the output channels that you want to make available to the addon. It will select the appropriate channel automatically and only report once per interrupt. Raid takes priority over Party which takes priority over Say.",
					order = 1
				},
				slashsay = {
					type = "toggle",
					name = "Say",
					width = "half",
					set = function(info, val) ezInterrupt.db.profile.slashSay = val ezInterrupt:UpdateOutputChannel() end,
					get = function(info) return ezInterrupt.db.profile.slashSay end,
					order = 4
				},
				slashparty = {
					type = "toggle",
					name = "Party",
					width = "half",
					set = function(info, val) ezInterrupt.db.profile.slashParty = val ezInterrupt:UpdateOutputChannel() end,
					get = function(info) return ezInterrupt.db.profile.slashParty end,
					order = 3
				},
				slashraid = {
					type = "toggle",
					name = "Raid",
					width = "half",
					set = function(info, val) ezInterrupt.db.profile.slashRaid = val ezInterrupt:UpdateOutputChannel() end,
					get = function(info) return ezInterrupt.db.profile.slashRaid end,
					order = 2
				},
				header2 = {
					type = "description",
					name = "In addition to the above, the addon can also whisper another player when you interrupt or show your interrupts in a custom channel.",
					order = 5
				},
				whispering = {
					type = "toggle",
					name = "Enable Whispering",
					set = function(info, val) ezInterrupt.db.profile.enableWhispering = val end,
					get = function(info) return ezInterrupt.db.profile.enableWhispering end,
					order = 6
				},
				whispertarget = {
					type = "input",
					name = "Which Player?",
					desc = "Input the name of the player you want to whisper to.",
					set = function(info, val) if val == "" then ezInterrupt.db.profile.whisperTarget = nil else ezInterrupt.db.profile.whisperTarget = val end end,
					get = function(info) return ezInterrupt.db.profile.whisperTarget end,
					order = 7
				},
				customchannel = {
					type = "toggle",
					name = "Enable Custom Channel",
					set = function(info, val) ezInterrupt.db.profile.enableCustomChannel = val end,
					get = function(info) return ezInterrupt.db.profile.enableCustomChannel end,
					order = 8
				},
				customchannelfield = {
					type = "input",
					name = "Which Channel?",
					desc = "Insert the number of the custom channel you want to report to. Example: 5",
					set = function(info, val) if val == "" then ezInterrupt.db.profile.customChannel = nil else ezInterrupt.db.profile.customChannel = val end end,
					get = function(info) return ezInterrupt.db.profile.customChannel end,
					order = 9
				}
			}
		},
		usewhere = {
			type = "group",
			name = "Where do you want to run ezInterrupt?",
			order = 6,
			inline = true,
			args = {
				useinparty = {
					type = "toggle",
					name = "In party instances",
					width = "full",
					set = function(info, val) ezInterrupt.db.profile.useInParty = val ezInterrupt:UpdateCombatLogParsing() end,
					get = function(info) return ezInterrupt.db.profile.useInParty end,
					order = 3,
				},
				useinraid = {
					type = "toggle",
					name = "In raid instances",
					width = "full",					
					set = function(info, val) ezInterrupt.db.profile.useInRaid = val ezInterrupt:UpdateCombatLogParsing() end,
					get = function(info) return ezInterrupt.db.profile.useInRaid end,
					order = 2
				},
				usealways = {
					type = "toggle",
					name = "Everywhere",
					width = "full",
					set = function(info, val) ezInterrupt.db.profile.useAlways = val ezInterrupt:UpdateCombatLogParsing() end,
					get = function(info) return ezInterrupt.db.profile.useAlways end,
					order = 1
				},						
			},
		},
		warningoptions = {
			type = "group",
			name = "Warning Options",
			order = 7,
			inline = true,
			args = {
				enablesound = {
					type = "toggle",
					name = "Play Sound",
					width = "full",
					set = function(info, val) ezInterrupt.db.profile.enableWarningSound = val end,
					get = function(info) return ezInterrupt.db.profile.enableWarningSound end,
					order = 1,
				},
				enablevisual = {
					type = "toggle",
					name = "Show Text (not yet implemented)",
					width = "full",					
					set = function(info, val) ezInterrupt.db.profile.enableWarningText = val end,
					get = function(info) return ezInterrupt.db.profile.enableWarningText end,
					order = 2
				},
				selectsound = {
					type = "select",
					name = "Which sound file?",
					style = "dropdown",
					width = "full",
					values = filterModeOptions,
					set = function(info, val) ezInterrupt.db.profile.filterMode = val end,
					get = function(info) return ezInterrupt.db.profile.filterMode end,
					order = 3
				},	
				filtering = {
					type = "select",
					name = "Spell filtering",
					style = "dropdown",
					width = "full",
					values = filterModeOptions,
					set = function(info, val) ezInterrupt.db.profile.filterMode = val end,
					get = function(info) return ezInterrupt.db.profile.filterMode end,
					order = 4
				},	
				blacklist = {
					type = "input",
					name = "Blacklist: add/remove",
					desc = "Add spell ID's to add them to the blacklist. Add again to remove them.",
					get = function(info) return nil end,
					set = function(info, val) UpdateFilterList(ezInterrupt.db.profile.blackList, val) end,
					order = 6
				},	
				whitelist = {
					type = "input",
					name = "Whitelist: add/remove",
					desc = "Add spell ID's to add them to the whitelist. Add again to remove them.",
					get = function(info) return nil end,
					set = function(info, val) UpdateFilterList(ezInterrupt.db.profile.whiteList, val) end,
					order = 9
				},	
				displayblacklist = {
					type = "execute",
					name = "Display Blacklist",					
					func = showBlacklistItems,
					order = 7
				},		
				displaywhitelist = {
					type = "execute",
					name = "Display Whitelist",					
					func = showWhitelistItems,
					order = 10
				},					
			}
		}
	}
}

function ezInterrupt:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ezInterruptDB", defaults, "Default");
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ezInterrupt", options);
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ezInterrupt", "ezInterrupt");
	self.RegisterChatCommand("ezinterrupt", "ChatCommand");
	self.RegisterChatCommand("ezi", "ChatCommand");
	init = true;
	-- set the playerInterrupts table used to determine failed interrupts
	local class = UnitClass("player");	
	if class == "Warrior" then
		tinsert(playerInterrupts, 6552); -- Pummel
		tinsert(playerInterrupts, 72); -- Shield Bash		
	elseif class == "Druid" then
		tinsert(playerInterrupts, 80964); -- Skull Bash (Bear)
		tinsert(playerInterrupts, 80965); -- Skull Bash (Cat)
	elseif class == "Rogue" then
		tinsert(playerInterrupts, 1766); -- Kick
	elseif class == "Paladin" then
		tinsert(playerInterrupts, 96231); -- Rebuke
	elseif class == "Deathknight" then
		tinsert(playerInterrupts, 47528); -- Mind Freeze 
	elseif class == "Shaman" then
		tinsert(playerInterrupts, 57994); -- Wind Shear
	elseif class == "Mage" then
		tinsert(playerInterrupts, 2139); -- Counterspell
	elseif class == "Warlock" then
		tinsert(playerInterrupts, 19647); -- Spell Lock
	elseif class == "Hunter" then
		tinsert(playerInterrupts, 34490); -- Silencing Shot
		tinsert(playerInterrupts, 50479); -- Nether Shock (Nether Ray pets)
		tinsert(playerInterrupts, 26090); -- Pummel (Gorilla pets)
	elseif class == "Priest" then
		tinsert(playerInterrupts, 15487); -- Silence
	end	
end

-- debugging: chat commands are not working!
function ezInterrupt:ChatCommand(input)
    if not input or input:trim() == "" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub("AceConfigCmd-3.0").HandleCommand(ezInterrupt, "ezi", "ezInterrupt", input)
    end
end

function ezInterrupt:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	if init then
		init = false;
	else
		self:UpdateOutputChannel(); 
		self:UpdateCombatLogParsing()
	end
end
function ezInterrupt:OnDisable()
	self:UnregisterEvent("PLAYER_ENTERING_WORLD");
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
end

function ValidateSpellID(spellID)
	if spellID == nil then
		return false
	elseif select(1, GetSpellInfo(spellID)) == nil then
		print("ezInterrupt: invalid spell ID")
		return false
	else
		print("ezInterrupt: added " .. select(1, GetSpellInfo(spellID)))
		return true
	end
end

function ezInterrupt:PLAYER_ENTERING_WORLD(event)
	self:UpdateOutputChannel(); -- sets the output channel according to player's instance and configuration settings
	self:UpdateCombatLogParsing(); -- enables or disables combat log parsing according to the player's configuration settings
end

function ezInterrupt:UpdateOutputChannel() -- sets OutputChannel to whatever channel is available depending on instance and configuration settings
	local instanceType = select(2, IsInInstance());
	if instanceType == "RAID" then
		if ezInterrupt.db.profile.slashRaid then outputChannel = "RAID";
		elseif ezInterrupt.db.profile.slashParty then outputChannel = "PARTY";
		elseif ezInterrupt.db.profile.slashSay then outputChannel = "SAY";
		else outputChannel = nil;
		end
	elseif instanceType == "PARTY" then
		if ezInterrupt.db.profile.slashParty then outputChannel = "PARTY";
		elseif ezInterrupt.db.profile.slashSay then outputChannel = "SAY";
		else outputChannel = nil;
		end
	elseif ezInterrupt.db.profile.slashSay then outputChannel = "SAY";
	else outputChannel = nil;
	end
end

function ezInterrupt:UpdateCombatLogParsing() -- enables/disables combat log parsing according to instance type and configuration settings
	local instanceType = select(2, IsInInstance());
	if ( instanceType == "RAID" and ezInterrupt.db.profile.useInRaid ) or ( instanceType == "PARTY" and ezInterrupt.db.profile.useInParty ) or ezInterrupt.db.profile.useAlways then
		ezInterrupt:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	else
		ezInterrupt:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	end
end

function ezInterrupt:InterruptReady() -- checks if the player has interrupt abilities that are not on cooldown. 
	local class = UnitClass("player");	
	if class == "Warrior" then
		local available1, norage1 = IsUsableSpell("Pummel");
		local cooldown1 = select(2, GetSpellCooldown("Pummel")); -- cooldown1 will be 0 if Pummel is not on cooldown
		local available2, norage2 = IsUsableSpell("Shield Bash"); 
		local cooldown2 = select(2, GetSpellCooldown("Shield Bash"));
		if ((available1 or norage1) and cooldown1 == 0) or ((available2 or norage2) and cooldown2 == 0) then -- IsUseableSpell("spell") returns arg1 as false and arg2 as true when the player has the ability but not enough rage/mana to use it, but we still want to count this as the interrupt being ready
			return true;
		else
			return false;
		end
	elseif class == "Druid" then
		local available1, noenergy = IsUsableSpell(80965); -- Skull Bash (Cat)
		local cooldown1 = select(2, GetSpellCooldown(80965)); -- Skull Bash (Cat)
		local available2, norage = IsUsableSpell(80964); -- Skull Bash (Bear)
		local cooldown2 = select(2, GetSpellCooldown(80964)); -- Skull Bash (Bear)
		if ((available1 or noenergy) and cooldown1 == 0) or ((available2 or norage) and cooldown2 == 0) then
			return true;
		else
			return false;
		end
	elseif class == "Rogue" then
		local cooldown = select(2, GetSpellCooldown("Kick"));
		if cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Paladin" then
		local cooldown = select(2, GetSpellCooldown("Rebuke"));
		if cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Deathknight" then
		local cooldown = select(2, GetSpellCooldown("Mind Freeze"));
		if cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Shaman" then
		local cooldown = select(2, GetSpellCooldown("Wind Shear"));
		if cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Mage" then
		local cooldown = select(2, GetSpellCooldown("Counterspell"));
		if cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Warlock" then
		local available, nomana = IsUsableSpell("Spell Lock");
		local cooldown = select(2, GetSpellCooldown("Spell Lock"));
		if (available or nomana) and cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Priest" then
		local available, nomana = IsUsableSpell("Silence");
		local cooldown = select(2, GetSpellCooldown("Spell Lock"));
		if (available or nomana) and cooldown == 0 then
			return true;
		else
			return false;
		end
	elseif class == "Hunter" then
		local available1, nofocus1 = IsUsableSpell("Silencing Shot");
		local cooldown1 = select(2, GetSpellCooldown("Silencing Shot"));
		local available2, nofocus2 = IsUsableSpell(26090); -- Pummel (Gorilla pets)
		local cooldown2 = select(2, GetSpellCooldown(26090)); -- Pummel (Gorilla pets)
		local available3, nofocus3 = IsUsableSpell(50479); -- Nether Shock (Nether Ray pets)
		local cooldown3 = select(2, GetSpellCooldown(50479)); -- Nether Shock (Nether Ray pets)
		if ((available1 or nofocus1) and cooldown1 == 0) or ((available2 or nofocus2) and cooldown2 == 0) or ((available3 or nofocus3) and cooldown3 == 0)then
			return true;
		else
			return false;
		end
	end
end

-- combat log parser
-- arguments passed are as follows:
-- 1. self, 2. timestamp, 3. eventType, 4. sourceGUID, 5. sourceName, 6. sourceFlags, 7. destGUID, 8. destName, 9. destFlags, 10. spellID, 11. spellName, 12. spellSchool, 13. extraSkillID, ...
function ezInterrupt:COMBAT_LOG_EVENT_UNFILTERED(_ , _, eventType, sourceGUID, sourceName, _, destGUID, destName, _, spellID, spellName, _, extraSkillID, ...)
	if self.db.profile.enableWarning and eventType == "SPELL_CAST_START" and sourceGUID == UnitGUID("target") and UnitCanAttack("player", "target") then -- checks for spell casts made by the current target, if the player can attack it
		if self.db.profile.filterMode == 2 then -- filterMode: 2 is "don't show spells in blackList"
			if tContains(self.db.profile.blackList, spellID) then 
				return
			end
		elseif self.db.profile.filterMode == 3 then -- filterMode: 3 is "show only spells in whiteList"
			if not tContains(self.db.profile.whiteList, spellID) then 
				return
			end
		end
		local unInterruptible = select(9, UnitCastingInfo("target")) -- checks if the cast is uninterruptible. 
		-- unInterruptible can be nil under certain circumstances. why?
		if unInterruptible ~= true and self:InterruptReady() then -- optimization: if both sound and text warnings are disabled, then enableWarning should also not be true
			if self.db.profile.enableWarningSound then 
				PlaySoundFile(self.db.profile.soundFile, "MASTER");
			end
			if self.db.profile.enableWarningText then	
				-- not yet implemented
			end
		end
		
	elseif self.db.profile.enableAnnouncing and eventType == "SPELL_INTERRUPT" and sourceGUID == UnitGUID("player") then
		-- optimization: add preformatted message
		local spell = GetSpellLink(extraSkillID)
		if self.db.profile.enableWhispering and self.db.profile.whisperTarget ~= nil then SendChatMessage("Interrupted " .. spell, "WHISPER", nil, self.db.profile.whisperTarget ); end
		if self.db.profile.enableCustomChannel and self.db.profile.customChannel ~= nil then SendChatMessage("Interrupted " .. spell, "CHANNEL", nil, self.db.profile.customChannel ); end	
		if outputChannel ~= nil then SendChatMessage("Interrupted " .. spell, outputChannel); end	-- optimization: outputChannel is only nil if all channels have been disabled, which should automatically disable all types of announcing.	
		
	elseif self.db.profile.enableAnnouncingForFailedInterrupts and eventType == "SPELL_MISSED" and sourceGUID == UnitGUID("player") and tContains(playerInterrupts, spellID) then
		if outputChannel ~= nil then SendChatMessage(spellName .. " failed! (" .. extraSkillID .. ")", outputChannel); end -- optimization: outputChannel is only nil if all channels have been disabled, which should automatically disable all types of announcing.
	end
end